// Gokurakuシリーズ共通の関数ライブラリー
// measure, merge, visualizeで完全同一

//// 定数
final String Version          = "3.1";
final String JapaneseFont     = "Meiryo";
final int    Delay_message    = 2000;          // ショートメッセージの表示時間（ミリ秒）

File _fp = new File("");
final String Separator = _fp.separator;  // システム依存のパス区切り子 (UNIX: /, Windwos: \\)

// グローバル変数
ShortMessage shortMessage;
InputInteger inputInteger;
InputString  inputString;
KeyStatus    keyStatus;

//// 変換関係
// 浮動小数点を小数点以下n桁で切り捨てる関数
float floorN(float f, int n) {
  return float(floor(f * pow(10, n))) / pow(10, n);
}

// RGBの色情報を文字列に変換する関数
String color2str(int r, int g, int b) {
  String s;
  s = "[" + r + "," + g + "," + b + "]";
  return s;
}

//// 文字列関係の関数
// StringList版
final String [] extensionListOfImage = {"jpg", "jpeg", "gif", "tga", "png"};  // TIFFはsaveFrame()できるが，loadImage()はできない．
final String [] extensionListOfText  = {"txt"};                               // 読み込めるテキストファイルの拡張子リスト

// File [] 内の全ファイル名を比較してソーティング（昇順）を行うクラス
class SortingFileList {
  StringList fList;  // 一時的な利用（ソートとリストからの削除）
  String     path;   // フルパス名（親ディレクトリ）
  // コンストラクタ
  // Fileのリストを受け取り，StringListに格納
  SortingFileList(File [] fs, String [] extList) {
    fList = new StringList();
    path = fs[0].getParent();
    // FileのリストをStringListに移し替える
    for (int i = 0; i < fs.length; i++) {
      if (fs[i].isDirectory()) continue;  // フォルダーはスキップ
      fList.append(fs[i].getName());      // ファイル名だけをStringListに格納（後でpathを追加
    }
    // 指定された拡張子のリスト以外のファイルをStringListから削除
    for (int i = fList.size() - 1; i >= 0; i--) {
      boolean flag = false;
      for (int j = 0; j < extList.length; j++) {
        String ext;
        String fname;
        fname = fList.get(i);
        ext = getExtension(fList.get(i)).toUpperCase(); 
        if (ext.equals(extList[j].toUpperCase())) {
          flag = true;
          break;
        }
      }
      if (flag == false) fList.remove(i);
      else               fList.set(i, path + Separator + fList.get(i));
    }
  }
  // 右から拡張子を取得する
  String getExtension(String str) {
    if (str.indexOf(".") == -1)               return "";
    if (str.indexOf(".") == str.length() - 1) return "";
    for (int i = str.length() - 1; i >= 0; i--) {
      if (str.charAt(i) == '.')               return str.substring(i + 1); 
    }
    return "";
  }
  // ソートして，フルパス名に戻した上でFileのリストに戻す
  File [] sort() {
    // ソート
    fList.sort();
    // Fileのリストに戻す
    File [] f = {};
    // Fileのリストに戻す
    f = (File [])expand(f, fList.size());
    for (int i = 0; i < fList.size(); i++) f[i] = new File(fList.get(i));
    return f;
  }
  // ファイルリストのコンソールへの出力：デバッグ用
  void showFileList() {
    for (int i = 0; i < fList.size(); i++) {
      println(i + ":" + fList.get(i));
    }
  }
}

// 文字列から数字を得て比較するクラス
//（merge中でメモリー内のデータを画像ファイル名や日時でソーティングするために用いる）
// 文字列はファイル名も可
// ANK文字列+数字列.拡張子を想定
// ピリオドは１つのみ（sample.jpg.txtはＮＧ）
// 数字列は0001のように左に0があっても構わない
class CompareString {
  String number1, number2;
  CompareString() {
    number1 = "";
    number2 = "";
  }
  // 比較（str1が大きければ 1，小さければ -1，等しければ 0 を返す）
  int compare(String str1, String str2) {
    // 念のための前処理
    if (str1 == null) str1 = "";
    if (str2 == null) str2 = "";
    // 正規化
    str1 = removeExtension(str1);
    str2 = removeExtension(str2);
    str1 = getNumbers(str1);
    str2 = getNumbers(str2);
    str1 = trimLeft(str1);
    str2 = trimLeft(str2);
    // 自明な場合
    if (str1.equals(str2))             return  0;  // 両方共 "" の場合を含む
    if (str1.length() > str2.length()) return  1;  // str2が "" の場合を含む
    if (str1.length() < str2.length()) return -1;  // str1が "" の場合を含む
    // 桁数が等しい場合
    for (int i = 0; i < str1.length(); i++) {
      int n1 = str1.charAt(i);
      int n2 = str2.charAt(i);
      if (n1 > n2) return  1;
      if (n1 < n2) return -1;
    }
    return 0;
  }
  // 右から拡張子を取得する
  String getExtension(String str) {
    if (str.indexOf(".") == -1)               return "";
    if (str.indexOf(".") == str.length() - 1) return "";
    for (int i = str.length() - 1; i >= 0; i--) {
      if (str.charAt(i) == '.')               return str.substring(i + 1); 
    }
    return "";
  }
  // 右から拡張子のみを取り除く
  String removeExtension(String str) {
    for (int i = str.length() - 1; i >= 0; i--) {
      if (str.charAt(i) == '.') return str.substring(0,i); 
    }
    return str;
  }
  // 右から数字のみの文字列を探して返す
  String getNumbers(String str) {
    String ret = "";
    char   c;
    for (int i = str.length() - 1; i >= 0; i--) {
      c = str.charAt(i);
      if ((c > '9') || (c < '0')) break;
      ret = c + ret;
    }
    return ret;
  }
  // 数字だけの文字列だとして，頭に付いている0を取り除く
  String trimLeft(String str) {
    for (int i = 0; i < str.length(); i++) {
      if (str.charAt(i) != '0') return str.substring(i);
    }
    return "0";
  }
};

//// ファイル関係のクラス
// CSV（タブ区切りの整数）のテキストを整数値に切り分けるクラス
class CsvStrings {
  String text;
  String str;
  float  value;
  // コンストラクター
  CsvStrings(String text) {
    this.set(text);
  }
  void set(String text) {
    this.text = text + "\t";
    value     =  float("NaN");  // floatの形式ではないのでNaNになる
  }
  // 次の値を取得する：正常に終了したらtrueを返し，valueに値が設定される
  boolean next() {
    if (text.indexOf("\t") == -1) return false;
    if (text.length() == 0)       return false;
    if (text.equals("\t"))        return false;
    str   = text.substring(0, text.indexOf("\t"));
    value = float(str);
    text  = text.substring(text.indexOf("\t") + 1);
    return true;
  }
  boolean is_float() {
    if (value != value) return false;  // NaN, ちょっとトリッキーな方法
    return true;
  }
}

// 指定されたデータファイルから１画像分のデータを取得してメモリーに保存するクラス
class ImportMeasuredFile {
  String import_folder;
  // コンストラクター
  ImportMeasuredFile() {
    import_folder = sketchPath();
  }
  
  // ファイル形式のチェック
  boolean validation(String fname) {
    File         fp;
    String[]     lines;  // データファイルから読み出した全テキストデータ

    // 存在するか？
    fp = new File(fname);
    if (  fp.isFile() != true) return false;
    // 読み込めるか？
    if ((lines = loadStrings(fname)) == null) return false;
    // 少なくとも２行以上か？
    if ((lines.length < 2)) return false;
    // １行目はscale, org_width, org_heightか？
    CsvStrings csv = new CsvStrings(lines[0]);
    csv.next();
    if (! csv.is_float() || csv.value <= 0.0)     return false;  // scale (float)
    csv.next();
    if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_width (int)
    csv.next();
    if (! csv.is_float() || int(csv.value) <= 0)  return false;  // org_height (int)
    // 互換性（Ver. 1.*のGokuraku_measureは画像ファイル名を記録していない）ので，この後に文字列があるかどうかはチェックしない
    // 互換性（Ver. 2.1以降のgookuraku_measureは計測日時を記録していないので，この後に文字列があるかどうかはチェックしない
    // ２行目は0で始まるか？
    csv.set(lines[1]); //<>//
    csv.next();
    if (! csv.is_float() || int(csv.value) != 0)  return false;

    return true;
  }
  
  // 実際にファイルのデータを読み込む
  OneData read(String fname) {
    return read(fname, false);
  }
  OneData read(String fname, boolean noConvert) {
    File         fp;
    CsvStrings   csv;
    OneData      onedata;
    String[]     lines;

    // ファイルチェック
    if (! validation(fname)) return null; 
    // ファイル読み込み
    if ((lines = loadStrings(fname)) == null) return null;

    // Scale, org_width, org_heightの読み出し
    onedata = new OneData();
    csv     = new CsvStrings(lines[0]);
    csv.next();
    onedata.scale      = csv.value;
    csv.next();
    onedata.org_width  = int(csv.value);
    csv.next();
    onedata.org_height = int(csv.value);
    csv.next();  // Ver.2 から，Gokuraku_maesureで画像ファイル名を記録するように変更
    if (csv.str.length() > 0) {
      onedata.filename      = csv.str;
    } else onedata.filename = "";
    csv.next();  // Ver.2.1 から，Gokuraku_maesureで計測日時を記録するように変更
    if (csv.str.length() > 0) {
      onedata.dateString    = csv.str;
    } else onedata.dateString = "";
    // Ver.3.0から計測データ数とスキップ数を記録するように変更したが，それは読み飛ばす（実際に読みこんだデータで再集計する）

    // データ本体の読み込み
    for (int i = 1; i < lines.length; i++) { //<>//
      int px, py, col_R, col_G, col_B;
      csv.set(lines[i]);
      csv.next();
      if (int(csv.value) != (i - 1)) {
        shortMessage.set_message("Error: データ形式が正しくない（No.が一致しない）");
        return null;
      }
      // X座標
      csv.next();
      px = int(csv.value);
      if (noConvert == false) int(px / (float)onedata.scale); 
      // Y座標
      csv.next();
      py = int(csv.value);
      if (noConvert == false) int(py / (float)onedata.scale);
      // R, G の空読み
      csv.next();
      col_R = int(csv.value);
      csv.next();
      col_G = int(csv.value);
      // B が存在するかどうかで一行のデータ数を確認
      if (csv.next() == false) {
        shortMessage.set_message("Error: データ形式が正しくない（一行のデータ数が少ない)");
        return null;
      }
      col_B = int(csv.value);
      onedata.add_point(px, py, col_R, col_G, col_B);
    }
    fp = new File(fname);  // ファイル名の分離と，エクスポートフォルダー未指定時に自動設定するため
    if (onedata.filename.length() == 0) onedata.filename = fp.getName();
    return onedata;
  }     
}

//// データに関するクラス
// 計測するデータに関するクラス
// １ファイル分（１計測点分）のデータ
class OneData {  // リスト構造
  String filename;
  String dateString;  // Ver.2.1から追加
  int    org_width, org_height;
  double scale;
  int    num;
  int [] px    = {};
  int [] py    = {};
  int [] col_R = {};
  int [] col_G = {};
  int [] col_B = {};
  
  // コンストラクター
  OneData() {
    filename   = "";
    org_width  = 0; 
    org_height = 0;
    scale      = 1.0;
    num        = 0;
  }
  // 座標データの追加（色は無視）
  void add_point(int _px, int _py) {
    add_point(_px, _py, 0, 0, 0);
  }
  // 座標データの追加
  void add_point(int _px, int _py, int _col_R, int _col_G, int _col_B) {
    px    = (int [])append(px, _px);
    py    = (int [])append(py, _py);
    col_R = (int [])append(col_R, _col_R);
    col_G = (int [])append(col_G, _col_G);
    col_B = (int [])append(col_B, _col_B);
    num   = px.length;
  }
  // 座標データの削除（measure専用）
  void delete_point() {
    px    = shorten(px);
    py    = shorten(py);
    col_R = shorten(col_R);
    col_G = shorten(col_G);
    col_B = shorten(col_B);
  }
  // 座標データのみ消去（measure専用：計測のRetry用）
  void points_clear() {
    num        = 0;
    px         = (int [])expand(px, 0);
    py         = (int [])expand(py, 0);
    col_R      = (int [])expand(col_R, 0);
    col_G      = (int [])expand(col_G, 0);
    col_B      = (int [])expand(col_B, 0);
  }
  // 全データ消去
  void all_clear() {
    filename   = "";
    org_width  = 0; 
    org_height = 0;
    points_clear();
    // scale = 1.0;  Ver.3.0からスケールはクリアしない仕様に変更
  }
  // 計測点の色情報を格納する（measure専用）
  void get_color(PImage img, int x, int y) {
    if (num <= 0) return;
    int idx = x + y * img.width;
    if (idx >= img.pixels.length) return;
    int c = img.pixels[idx];
    col_R[num - 1] = int(red(c));
    col_G[num - 1] = int(green(c));
    col_B[num - 1] = int(blue(c));
  }
  String col2str(int n) {
    if (n >= num) return "";
    return color2str(col_R[n], col_G[n], col_B[n]); 
  }
  // 計測されたデータの中のスキップされた計測点の数を返す
  int countSkip() {
    int skip = 0;
    for (int i = 0; i < num; i++) {
      if (px[i] == -1) skip++;
    }
    return skip;
  }
}

//// プロットに関するクラス
// プロットの種類
final String [] plotName = {"none", "cross", "circle", "circlefill", "rect", "rectfill", "scope", "scopefill"}; 
// プロットを描画
class Plot {
  char mode;       // (0: none), 1: cross,  2: circle, 3: circlefill, 4: rect, 5: rectfill, 6, scope, 7. scopefill
  int  size;
  int  p_color;    // 0, 1, 2
  int  lineWidth;

  // コンストラクター
  Plot() {
    mode      = 1;  // default: cross
    size      = Plot_size;
    lineWidth = 1;
  }
  // 全プロットの表示
  void draw_plot(OneData data) {  // measureでは計測した座標にscaleが掛かっているので変換不要
    draw_plot(data, 1.0);
  }
  void draw_plot(OneData data, float scale) {
    strokeWeight(lineWidth);
    for (int i = 0; i < data.num; i++) {
      int x = data.px[i];
      int y = data.py[i];
      if ((x == -1) || (y == -1)) continue; // スキップされた計測点への対応

      if (ci.enable_scale && (scale != 1.0)) {
        x = (int)(x * scale);
        y = (int)(y * scale);
      }
      plot(x, y);
    }
    strokeWeight(1);
  }
  // プロットの形状を変更
  void change_shape() {
    if (++mode >= plotName.length) mode = 1;
  }
  String get_shape() {
    return plotName[mode];
  }
  // プロットの色を変更
  void change_color() {
    if (++p_color > Plot_color.length - 1) p_color = 0; 
  }
  // プロットを描く
  void plot(int x, int y) {
    stroke(this.red(), this.green(), this.blue());
    noFill();
    switch (mode) {
      case 0 : break;
      case 1 : // cross
        cross(x, y);
        break;
      case 3 : // circlefill
        fill(this.red(), this.green(), this.blue());
      case 2 : // circle 
        circle(x, y);
        break;
      case 5 : // rectfill
        fill(this.red(), this.green(), this.blue());
      case 4 : // rect 
        rectangle(x, y); 
        break;
      case 7 : // scopefill
        fill(this.red(), this.green(), this.blue());
      case 6 : // scope      
        cross(x, y, true);
        circle(x, y);
        break;
      default : mode = 0;
    }
  }
  // 十文字
  void cross(int x, int y) {
    cross(x, y, false);
  }
  // 十文字（照準マーク対応）
  void cross(int x, int y, boolean large) {
    int s = size;
    if (large) s *= 2;
    line(x - s, y, x + s, y);
    line(x, y - s, x, y + s);
  }
  // 円
  void circle(int x, int y) {
    ellipse(x, y, size * 2, size * 2);
  }
  // 四角
  void rectangle(int x, int y) {
    rectMode(CENTER); rect(x, y, size, size);
  }
  // 線分を描く（太さとカラーは事前に指定すること）
  void stick(int x0, int y0, int x1, int y1, float scale) {
    if (scale != 1.0) {
      x0 = (int)(x0 * scale);
      x1 = (int)(x1 * scale);
      y0 = (int)(y0 * scale);
      y1 = (int)(y1 * scale);
    }
    noFill();
    line(x0, y0, x1, y1);
  }
  // 色情報のテキスト化
  String color_str() {
    return color2str(this.red(), this.green(), this.blue());
  }
  int red() {
    return Plot_color[p_color][0];
  }
  int green() {
    return Plot_color[p_color][1];
  }
  int blue() {
    return Plot_color[p_color][2];
  }
}

// 線分描画のクラス
class Stick {
  boolean     enable_stick;  // 線分表示を有効化するか
  int [][]    define_stick;  // Stick picture(line segment)の定義

  // コンストラクター
  Stick() {
    define_stick = Define_stick;
    enable_stick = false;
  }
  // 線分定義領域のクリア
  void clearDefineStick() {
    define_stick = (int [][])expand(define_stick, 0);
  }
  // 線分定義ファイルの読み込み
  void readDefineStick(String fname) {
    File      fp;
    String [] lines;
    int [][]  def_stick = {};

    // メモリーへの読み込み
    if ((lines = loadStrings(fname)) == null) {
      shortMessage.set_message("Can't read define_stick file");
      return;
    }
    // 書式のチェック
    for (int i = 0; i < lines.length; i++) {
      CsvStrings csv = new CsvStrings(lines[i]);
      for (int j = 0; j < 6; j++) {
        if (!csv.next() || ! csv.is_float() || int(csv.value) < 0)  {  // start, end, R, G, B, Width
          shortMessage.set_message("Error: format error (not integer or negative)");
          return;
        }
      }
      if (csv.next()) {
        shortMessage.set_message("Error: format error (too much, 6 parameters)");
        return;  // 7項目以上あったらエラー
      }
    }
    // 格納
    for (int i = 0; i < lines.length; i++) {
      int [] onedef = {0, 0, 0, 0, 0, 0};
      CsvStrings csv = new CsvStrings(lines[i]);
      for (int j = 0; j < 6; j++) {
        csv.next();
        onedef[j] = int(csv.value);
      }
      if ((onedef[0] < 0) || (onedef[1] < 0)) {
        shortMessage.set_message("Error: format error (Start and End must be large equal 0)");
        return;
      }
      if ((onedef[2] > 255) || (onedef[3] > 255) || (onedef[4] > 255)) {
        shortMessage.set_message("Error: format error (color code must be 0 - 255)");
        return;
      }
      def_stick = (int [][])append(def_stick, onedef);
    }
    define_stick = def_stick;
    shortMessage.set_message("Imported define_stick file :" + define_stick.length + " data");
    enable_stick = true;  // 読み込みに成功したら自動的に有効にする
  }
  // 線分描画（measure専用）
  void draw_stick(OneData data) { // measureでは計測した座標にscaleが掛かっているので変換不要
    draw_stick(data, 1.0);
  }
  // 線分描画
  void draw_stick(OneData data, float scale) {
    if (enable_stick) {
      int x0, y0, x1, y1;
      // 未定義あるいは読み込み失敗時（０から順番に線で結ぶ）
      if (define_stick.length == 0) {
        stroke(ci.plotMark.red(), ci.plotMark.green(), ci.plotMark.blue());
        strokeWeight(ci.plotMark.lineWidth);
        for (int i = 0; i < data.num - 1; i++) {
          x0 = data.px[i]; y0 = data.py[i]; x1 = data.px[i+1]; y1 = data.py[i+1];
          if ((x0 == -1) || (y0 == -1) || (x1 == -1) || (y1 == -1)) continue;  // スキップされた計測点への対応
          if (ci.enable_scale && (scale != 1.0)) ci.plotMark.stick(x0, y0, x1, y1, scale);
          else                                   ci.plotMark.stick(x0, y0, x1, y1, 1.0);
        }
      }else {
        // 定義済みあるいは読み込み成功時
        for (int i = 0; i < define_stick.length; i++) {
          int start, end, r, g, b, weight;
          start  = define_stick[i][0];
          end    = define_stick[i][1];
          r      = define_stick[i][2];
          g      = define_stick[i][3];
          b      = define_stick[i][4];
          weight = define_stick[i][5];
          if ((start >= data.num) || (end >= data.num)) continue;

          x0 = data.px[start];
          y0 = data.py[start];
          x1 = data.px[end];
          y1 = data.py[end];
          if ((x0 == -1) || (y1 == -1) || (x1 == -1) || (y1 == -1)) continue;  // スキップされた計測点への対応

          stroke(r, g, b);
          strokeWeight(weight);
          if (ci.enable_scale && (scale != 1.0)) {
                 ci.plotMark.stick(x0, y0, x1, y1, scale);
          } else ci.plotMark.stick(x0, y0, x1, y1, 1.0);
        }
      }
      strokeWeight(1);
    }
  }
  // 線分表示（stick picture）の無効／有効切り替え
  void switch_enable_stick() {
    enable_stick = (enable_stick ? false : true);
  }
}

//// メニュー関係のユーザーインターフェース
// 色の列挙型の宣言
enum TextColor {normal, option, input};
// メニュー画面の記述
class MenuScreen {
  int x,  y;
  int dx, dy;
  // コンストラクター
  MenuScreen(int x0, int y0, int dx0, int dy0) {
    x  = x0;
    y  = y0;
    dx = dx0;
    dy = dy0;
  }
  // タイトル表示
  void display_title(String title, int r, int g, int b) {
    fill(r, g, b);
    textSize(28);
    text(title + " (@tatsuva) Ver." + Version, x, y);
    y += dy;
  }
  // タイトル下のサブタイトル表示
  void display_subtitle(String subtitle) {
    textSize(18);
    text(subtitle, x, y);
    y += dy;
  }
  // メニュー項目の色設定
  void text_color(TextColor col) {
    fill(255, 80, 80);
    if (col == TextColor.option) fill(255, 255, 0);
    if (col == TextColor.input)  fill(255, 255, 255);
  }
  // メニュー項目
  void display_menu(String keyStr, String description) {
    textSize(24);
    text(keyStr,             x,      y);
    text(": " + description, x + dx, y);
    y += dy;
  }
  // キー入力プロンプト
  void display_keyInputPrompt(String prompt) {
    fill(255, 80, 80);
    textSize(24);
    if      (inputInteger.inputing_now) text(prompt + " = " + inputInteger.value, x, y); 
    else if (inputString.inputing_now)  text(prompt + " = " + inputString.value,  x, y); 
    y += dy;
  }
  // ノート表示
  void display_note(String note) {
    textSize(16);
    fill(255, 255, 255);
    text(note, x, y);
    y += 20;
  }
}

// 正の整数をキーボードから入力するためのクラス
class InputInteger {
  boolean inputing_now;  // 入力中（未確定）
  boolean halt;          // 中断（確定）
  int     value;
  
  // コンストラクター
  InputInteger() {
    inputing_now = false;
    halt         = false;
    value        = 0;
  }
  // 入力開始
  void start() {
    start(0);
  }
  void start(int init) {
    inputing_now = true;
    halt         = false;
    value        = init;
  }
  // 終了
  void end() {
    inputing_now = false;
    if (value < 0) value = 0;
  }
  // キーが入力された場合
  boolean keyin(char c) {
    int code;  // キャラクターコード
    code = int(c) - 48;  // 0 - 9
    if ((code >= 0) && (code <= 9)) {
      value = value * 10 + code;
    } else if (c == BACKSPACE) {
      value /= 10;
    } else if (c == ENTER) {
      end();
    } else if (c == ' ') {
      value = 0;
    } else if (c == ESC) {
      halt = true;
      end();
      shortMessage.set_message("Canceled");
    }
    return inputing_now;
  }
}

// 文字列をキーボードから入力するためのクラス
class InputString {
  boolean inputing_now;  // 入力中（未確定）
  boolean halt;          // 中断（確定）
  String  value;
  
  // コンストラクター
  InputString() {
    inputing_now = false;
    halt         = false;
    value        = "";
  }
  // 入力開始
  void start() {
    start("");
  }
  void start(String init) {
    inputing_now = true;
    halt         = false;
    value        = init;
  }
  // 終了
  void end() {
    inputing_now = false;
  }
  // キーが入力された場合
  boolean keyin(char c) {
    boolean err = true;
    if ((c >= 'a') && (c <= 'z'))      err = false;
    else if ((c >= 'A') && (c <= 'Z')) err = false;
    else if ((c >= '0') && (c <= '9')) err = false;
    else if ((c == '-') || (c == '(') || (c == ')')) err = false;
    if (! err) value += c;
    else {
      if (c == BACKSPACE) {
        if (value.length() > 0) value = value.substring(0, value.length() - 1);
      } else if (c == ENTER) {
        end();
      } else if (c == ' ') {
        value = "";
      } else if (c == ESC) {
        halt = true;
        end();
        shortMessage.set_message("Canceled");
      }
    }
    return inputing_now;
  }
}

// キーの同時押しに対応するためのクラス
class KeyStatus {
  boolean left, right, shift, ctrl;
  KeyStatus() {
    left = right = shift = ctrl = false;
  }
  void push(int kc) {
    switch (kc) {
      case LEFT    : left  = true; break;
      case RIGHT   : right = true; break;
      case SHIFT   : shift = true; break;
      case CONTROL : ctrl  = true; break;
    }
  }
  void release(int kc) {
    switch (kc) {
      case LEFT    : left  = false; break;
      case RIGHT   : right = false; break;
      case SHIFT   : shift = false; break;
      case CONTROL : ctrl = false; break;
    }
  }
};

// ウインドウとコンソールエリアに短いメッセージを一時的に表示するためのクラス
class ShortMessage {
private  int    start;
private  String message;
  ShortMessage() {
    start   = millis();
    message = "";
  }
  void clear() {
    message = "";
  }
  void set_message(String str) {
    set_message(str, false);
  }
  void set_message(String str, boolean noconsole) {
    message = str;
    if (! noconsole) println(str);
    start   = millis();
  }
  void display_short_message(int x, int y) {
    if (millis() - start < Delay_message) {
      fill(200, 200, 200);
      textSize(16);
      text(message, x, y);
    }
  }
}

//// D&D関係
// 参考URL（ドラッグ＆ドロップに関して
// https://qiita.com/rain-lotus/items/258fa3defd41846bfd50
// https://sites.google.com/site/gutugutu30/other/processingdedoragguandodoroppu

import java.awt.datatransfer.*;  
import java.awt.dnd.*;  
import java.io.File;  
import java.io.IOException;  
import java.awt.Component;
import java.util.List;

DropTarget dropTarget;
Component  component;

void drop_init() {  
  component = (Component)this.surface.getNative();
  dropTarget = new DropTarget(component, new DropTargetListener() {
    public void dragEnter(DropTargetDragEvent dtde) {
    }  
    public void dragOver(DropTargetDragEvent dtde) {
    }  
    public void dropActionChanged(DropTargetDragEvent dtde) {
    }  
    public void dragExit(DropTargetEvent dte) {
    }  
    public void drop(DropTargetDropEvent dtde) {  
      dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);  
      Transferable trans = dtde.getTransferable();  
      List<File> fileNameList = null;  
      if (trans.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {  
        try {  
          fileNameList = (List<File>)  
            trans.getTransferData(DataFlavor.javaFileListFlavor);
        } 
        catch (UnsupportedFlavorException ex) {
        } 
        catch (IOException ex) {
        }
      }  
      if (fileNameList == null) return;  
      fileSelected(fileNameList);
    }
  }
  );
}
